#include <Wire.h>
#include <LSM303.h>

LSM303 compass;     //Create the data structures

char nmea[50+1];    //Space for the NMEA sentence

void setup() {
  Serial.begin(38400);  //NMEA 0183 High Speed, On a Pololu A-Star 32U4 this is the USB, so the numerical value is not relevant
  Serial1.begin(38400); //On a A-Star 32U4 this is the serial port on pins 0 (RX) and 1 (TX)
  Wire.begin();         //Initialize the I2C communications  
  compass.init();
  compass.enableDefault();
  byte t = compass.readReg( LSM303::CTRL5 );  
  compass.writeReg( LSM303::CTRL5, t | 0x80 );  //Turn on the temperature sensor
  
  /*
  Calibration values; the default values of +/-32767 for each axis
  lead to an assumed magnetometer bias of 0. Use the Calibrate example
  program to determine appropriate values for your particular unit.
  */
  compass.m_min = (LSM303::vector<int16_t>){-3563, -3167, -3648};    //Note! You need to do your own calibration and replace these values
  compass.m_max = (LSM303::vector<int16_t>){+3364, +3409, +3322};
 }

float filtered_hdg( float hdg ) {  // Filter the Heading to make it less jumpy
  
  static float hdg_f = 0.0;  //Filtered heading
  float fac;                 //Moving average filter multiplier
  const float GAIN = 1/15.0; //The larger the gain, the more the most recent hdg change will affect the filtered result, fac = Delta degrees * GAIN
  const float FAC_MAX = 0.3; //The maximum modified moving average factor, the larger it is, the more the latest hdg can affect the result, range (0..1]
  
  /*
  First make sure we handle the zero crossing gracefully
  If the heading crosses the zero degree line from NW to NE we let the filter input be > 360 to avoid the abrupt step
  Likewise if the heading crosses the zero degree line  NE -> NW, we let the filter input be < 0
  */
  if ( abs( hdg - hdg_f ) > 180.0 ) {
    if ( hdg > hdg_f ) {
      hdg -= 360.0;
    }
    else {
      hdg += 360.0;
    }
  }
  fac = abs( hdg - hdg_f ) * GAIN;          // Speed up tracking for fast course changes
  if (fac > FAC_MAX ) {                     // Limit the max rate of change
    fac = FAC_MAX;
  } 
  hdg_f = hdg_f * ( 1 - fac ) + fac * hdg;  // Modified Moving Average filter
  
  // After the filter we may need to account for a zero degree line crossing as we have allowed < 0 and > 360 degree inputs to the filter
  
  if ( hdg_f  < 0.0 ) {            
    hdg_f += 360.0;
  }
  else {
    if ( hdg_f >= 360.0 ) {
      hdg_f -= 360.0;
    }  
  }
  return hdg_f;
}

void nmea_csum( char nmea[] ) {   //Adds *CS<cr><lf> to the end of the NMEA string

unsigned char csum; //NMEA checksum
char csumh[3+1];    //NMEA checksum string

  csum = 0;
  for ( int n=1; n < strlen(nmea); n++ ) {  //XOR of the characters between the $ and the *
    csum ^= (unsigned char)nmea[n];
  }
  snprintf( csumh, 4, "*%02X", csum );
  strcat( nmea, csumh );
  strcat( nmea, "\r\n" );
}

void nmea_hdg( char nmea[], float hdg ) {  // Construct the NMEA 0183 HDG sentence
  
  strcpy(nmea,"$HCHDG,");
  if ( hdg > 359.94 ) { //Prevent dtostrf() from rounding to 360.0 
    hdg = 0.0;
  }  
  dtostrf( hdg, 2, 1, &nmea[7] ); //Note: no leading zeros, some odd equipment might require them
  strcat( nmea, ",,,," );        //strcat( nmea, ",0.0,E,0.0,E" );  //This would be the full sentence with deviation and variation
  nmea_csum( nmea );
  
}

void nmea_mda( char nmea[], float temp ) {  // Construct the NMEA 0183 MDA (Meteorological Composite) sentence for outside temperature
  
  strcpy( nmea, "$HCMDA,,,,," );  // MDA is deprecated, but many devices do not recognize the alternative XDR (Transducer Measurement) sentence
  dtostrf( temp, 2, 1, &nmea[11] );
  strcat( nmea, ",C,,,,,,,,,,,,,," );
  nmea_csum( nmea );
}

void loop() {
  
  float  f_hdg;
  
  for (int n = 0; n < 6; n++ ) {
    compass.read();
  
    // Process magnetic heading
    f_hdg = filtered_hdg( compass.heading((LSM303::vector<int>){1, 0, 0}) );
    nmea_hdg( nmea, f_hdg ); 
    Serial.print( nmea );
    Serial1.print( nmea );
    delay( 155 );
  }
  
  // Process temperature
  byte tl = compass.readReg(LSM303::TEMP_OUT_L);
  byte th = compass.readReg(LSM303::TEMP_OUT_H);
  int temperature_raw = (int16_t)(th << 8 | tl);
  nmea_mda( nmea, 25.0 + (float)temperature_raw / 8 );
  Serial.print( nmea );
  Serial1.print( nmea );

}
